# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - Graphviz Parser
    Based loosely on GNUPLOT parser by MoinMoin:KwonChanYoung

    @copyright: 2008 Wayne Tucker
    @copyright: 2011, 2012 Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL, see COPYING for details.
"""

"""
    BASIC USAGE:

    embed a visualization of a graph in a wiki page:

{{{#!graphviz
digraph G {
    A -> B;
};
}}}

    ADVANCED USAGE:

    This parser will check the first lines of the Graphviz data for C++ style
    comments instructing it to use a different filter (dot, neato, twopi,
    circo, or fdp - see http://graphviz.org/ for more info), use a different
    format for the output (see the FORMATS list in the Parser class below),
    or to generate and pass a client-side image map.

    Options:
      filter - the filter to use (see Parser.FILTERS)
      format - the output format (see Parser.FORMATS)
      cmapx - the map name to use for the client-side image map.  Must match
              the graph name in the graph definition and shouldn't conflict
              with any other graphs that are used on the same page.

    embed a visualization of a graph in a wiki page, using the dot filter and
    providing a client-side image map (the filter=dot and format=png options are
    redundant since those are the defaults for this parser):

{{{#!graphviz
//filter=dot
//format=png
//cmapx=DocumentationMap
digraph DocumentationMap {
    FrontPage [href="FrontPage", root=true];
    HelpOnEditing [href="HelpOnEditing"];
    SyntaxReference [href="SyntaxReference"];
    WikiSandBox [href="WikiSandBox", color="grey"];
    MoinMoin [href="http://moinmo.in"];
    FrontPage -> WikiSandBox;
    FrontPage -> MoinMoin;
    WikiSandBox -> HelpOnEditing;
    WikiSandBox -> SyntaxReference;
    SyntaxReference -> FrontPage;
};
}}}


    KNOWN BUGS:
      - Hasn't been thoroughly checked for potential methods of injecting
        arbitrary HTML into the output.
      - Only compatible with HTML rendering
      - May not use all of the MoinMoin interfaces properly - this is a
        quick hack based on looking at an example and digging through the
        MoinMoin source.  The MoinMoin development docs haven't been
        consulted (yet).
      - Comments must start at the beginning of the graphviz block, and at the
        beginning of their respective lines.  They must also not contain
        any extra whitespace surrounding the = sign.

"""

# Change this to the directory that the Graphviz binaries (dot, neato, etc.)
# are installed in.

BINARY_PATH = '/usr/bin'

import os
import subprocess
import sha
import re

from MoinMoin import config
from MoinMoin.action import AttachFile
from MoinMoin import log
from MoinMoin import wikiutil

logging = log.getLogger(__name__)

class GraphVizError(RuntimeError):
    pass

Dependencies = []

class Parser:
    """Uses the Graphviz programs to create a visualization of a graph."""

    FILTERS = ['dot', 'neato', 'twopi', 'circo', 'fdp']
    IMAGE_FORMATS = ['png', 'gif']
    OBJECT_FORMATS = ['svg', 'svgz']
    OUTPUT_FORMATS = IMAGE_FORMATS + OBJECT_FORMATS
    FORMATS = OUTPUT_FORMATS + \
        ['ps', 'fig', 'mif', 'hpgl', 'pcl', 'dia', 'imap', 'cmapx']
    extensions = []
    Dependencies = Dependencies

    attach_regexp = re.compile(
        r"graphviz_"
        r"(?P<digest>.*?)"
        r"(?:"                              # begin optional section
        r"_(?P<width>.*?)_(?P<height>.*?)"  # dimensions
        r")?"                               # end optional section
        r"\.(?P<format>.*)"
        r"$")

    attr_regexp = re.compile(
        r"(?P<attr>width|height)"
        r"\s*=\s*"
        r"""(?P<quote>['"])"""              # start quote
        r"(?P<value>.*?)"
        r"""(?P=quote)""",                  # matching quote
        re.UNICODE)

    def __init__(self, raw, request, **kw):
        self.raw = raw
        self.request = request

    def format(self, formatter):

        "Using the 'formatter', return the formatted page output."

        request = self.request
        page = request.page
        _ = request.getText

        request.flush() # to identify error text

        self.filter = Parser.FILTERS[0]
        format = 'png'
        cmapx = None
        width = None
        height = None

        raw_lines = self.raw.splitlines()
        for l in raw_lines:
            if not l[0:2] == '//':
                break

            parts = l[2:].split("=")
            directive = parts[0]
            value = "=".join(parts[1:])

            if directive == 'filter':
                filter = value.lower()
                if filter in Parser.FILTERS:
                    self.filter = filter
                else:
                    logging.warn('unknown filter %s' % filter)

            elif directive == 'format':
                value = value.lower()
                if value in Parser.FORMATS:
                    format = value

            elif directive == 'cmapx':
                cmapx = wikiutil.escape(value)

        if not format in Parser.OUTPUT_FORMATS:
            raise NotImplementedError, "only formats %s are currently supported" % \
                Parser.OUTPUT_FORMATS

        if cmapx:
            if not format in Parser.IMAGE_FORMATS:
                logging.warn('format %s is incompatible with cmapx option' % format)
                cmapx = None

        digest = sha.new(self.raw).hexdigest()

        self.pagename = formatter.page.page_name
        self.attach_dir = AttachFile.getAttachDir(request, self.pagename, create=1)
        self.delete_old_graphs(formatter)

        attrs = self.find_graph(digest, format)
        if not attrs:
            attrs = self.graphviz(self.raw, digest, format)

        chart = self.get_chartname(digest, format, attrs)
        url = AttachFile.getAttachUrl(self.pagename, chart, request)

        if format in Parser.IMAGE_FORMATS:
            if cmapx:
                request.write('\n' + self.graphviz(self.raw, digest, "cmapx") + '\n')
                request.write(formatter.image(src="%s" % url, usemap="#%s" % cmapx, **self.get_format_attrs(attrs)))
            else:
                request.write(formatter.image(src="%s" % url, alt="graphviz image", **self.get_format_attrs(attrs)))
        else:
            request.write(formatter.transclusion(1, data=url, **self.get_format_attrs(attrs)))
            request.write(formatter.text(_("graphviz image")))
            request.write(formatter.transclusion(0))

    def find_graph(self, digest, format):

        "Find an existing graph using 'digest' and 'format'."

        attach_files = AttachFile._get_files(self.request, self.pagename)

        for chart in attach_files:
            match = self.attach_regexp.match(chart)

            if match and \
                match.group("digest") == digest and \
                match.group("format") == format:

                return match.groupdict()

        return None

    def get_chartname(self, digest, format, attrs):

        "Return the chart name for the 'digest', 'format' and 'attrs'."

        wh = self.get_dimensions(attrs)
        if wh:
            dimensions = "_%s_%s" % wh
        else:
            dimensions = ""
        return "graphviz_%s%s.%s" % (digest, dimensions, format)

    def delete_old_graphs(self, formatter):

        "Using the 'formatter' for page information, delete old graphs."

        page_info = formatter.page.lastEditInfo()
        try:
            page_date = page_info['time']
        except KeyError, ex:
            return

        attach_files = AttachFile._get_files(self.request, self.pagename)

        for chart in attach_files:
            match = self.attach_regexp.match(chart)

            if match and match.group("format") in Parser.FORMATS:
                fullpath = os.path.join(self.attach_dir, chart).encode(config.charset)
                st = os.stat(fullpath)
                chart_date = self.request.user.getFormattedDateTime(st.st_mtime)
                if chart_date < page_date:
                    os.remove(fullpath)

    def graphviz(self, graph_def, digest, format):

        "Using the 'graph_def' and 'digest', generate output in the given 'format'."

        p = subprocess.Popen(['%s/%s' % (BINARY_PATH, self.filter), '-T%s' % format], shell=False, \
                             stdin=subprocess.PIPE, \
                             stdout=subprocess.PIPE, \
                             stderr=subprocess.PIPE)

        p.stdin.write(graph_def)
        p.stdin.flush()
        p.stdin.close()

        p.wait()

        # Graph data always goes via standard output so that we can extract the
        # width and height if possible.

        output, attrs = self.process_output(p.stdout, format)
        errors = p.stderr.read()

        if len(errors) > 0:
            raise GraphVizError, errors

        # Copy to a file, returning the width and height if possible.

        if format != "cmapx":
            chart = self.get_chartname(digest, format, attrs)
            filename = os.path.join(self.attach_dir, chart).encode(config.charset)

            f = open(filename, "wb")
            try:
                f.write(output)
            finally:
                f.close()

            return attrs

        # Otherwise, return the output.

        else:
            return output

    def process_output(self, output, format):

        "Process graph 'output' in the given 'format'."

        # Return the raw output if SVG is not being produced.

        if format != "svg":
            return output.read(), {}

        # Otherwise, return the processed SVG output.

        processed = []
        found = False
        attrs = {}

        for line in output.xreadlines():
            if not found and line.startswith("<svg "):
                for match in self.attr_regexp.finditer(line):
                    attrs[match.group("attr")] = match.group("value")
                found = True
            processed.append(line)

        return "".join(processed), attrs

    def get_dimensions(self, attrs):

        "Return a (width, height) tuple using the 'attrs' dictionary."

        if attrs.has_key("width") and attrs.has_key("height"):
            return attrs["width"], attrs["height"]
        else:
            return None

    def get_format_attrs(self, attrs):

        "Return a dictionary based on 'attrs' with only formatting attributes."

        dattrs = {}
        for key in ("width", "height"):
            if attrs.has_key(key):
                dattrs[key] = attrs[key]
        return dattrs
